// Cocoa includes
#import <Foundation/Foundation.h>
#import <UIKit/UIKit.h>
#import <Preferences/PSListController.h>
#import <Preferences/PSSpecifier.h>

// standard and project-specific includes
#include <notify.h>
#include "common.h"


// handy definitions
#define LOCALIZE(textid,defaulttext) [[self bundle] localizedStringForKey: (textid) value: (defaulttext) table: nil]
#define BOOL_FOR_ID(id) ((CFBooleanRef) [self readPreferenceValue: [self specifierForID: (id)]] == kCFBooleanTrue ? true : false)
#define STRING_FOR_ID(id) [[self readPreferenceValue: [self specifierForID: (id)]] UTF8String]


@interface PCNetworking : PSListController
{
}

- (id) specifiers;
- (NSArray *) localizedSpecifiersForSpecifiers: (NSArray *) s;
- (id) navigationTitle;
- (id) getDaemonState: (PSSpecifier *) spec;
- (void) setConfigValue: (id) value forSpecifier: (PSSpecifier *) spec;
- (void) pushButton: (PSSpecifier *) spec;

@end


// samba share structure type definition
typedef struct sambashare_s
{
   bool is_enabled;
   char name[256];
   char path[256];
   char description[256];
   bool is_readonly;
   bool is_hidden;
} sambashare_t;


// samba settings structure type definition
typedef struct sambasettings_s
{
   bool is_running;
   char workgroup[256];
   char netbios_name[256];
   char description[256];
   char password[256];
   sambashare_t storage;
   sambashare_t camera;
   sambashare_t fsroot;
} sambasettings_t;


// global variables
static sambasettings_t samba;
static unsigned long device_id;
static unsigned long activation_code;
static bool is_activated;


// prototypes of local functions
static void ComputeDeviceIDAndActivationCode (const char *unique_id);
static bool Samba_IsDaemonRunning (void);
static void Samba_Setup (sambasettings_t *settings, bool is_running, const char *workgroup, const char *netbios_name, const char *description, const char *password);
static void Samba_SetupShare (sambashare_t *share, bool is_enabled, const char *name, const char *path, const char *description, bool is_readonly, bool is_hidden);
static void Samba_WriteConfig (void);


@implementation PCNetworking


- (id) specifiers
{
   // have the specifiers not been loaded yet ? If so, read them from the PList file
   if (!_specifiers)
   {
      // build the device ID and activation code first
      ComputeDeviceIDAndActivationCode ([[[UIDevice currentDevice] uniqueIdentifier] UTF8String]);

      // load the specifiers from the PList file
      _specifiers = [[self loadSpecifiersFromPlistName: @"PCNetworking" target: self] retain];

      // see if we're activated
      is_activated = ((unsigned long) atol (STRING_FOR_ID (@"activation_code")) == activation_code);
   }

   // localize the specifiers
   _specifiers = [self localizedSpecifiersForSpecifiers: _specifiers];
   return (_specifiers);
}


- (NSArray *) localizedSpecifiersForSpecifiers: (NSArray *) s
{
   // this function translates the preference bundle's specifiers data as read from the
   // BundleName.plist file into localized strings from the .lproj subdirectories, by
   // askin the localization API to do that replacement work for us.

   unsigned int specifier_index;
   unsigned int specifier_count;
   unsigned int key_index;
   unsigned int key_count;
   NSMutableDictionary *new_titles;
   NSString *key;
   id specifier;

   specifier_count = [s count]; // get the number of items in the specifiers array

   // cycle through each specifier in the preferences bundle...
   for (specifier_index = 0; specifier_index < specifier_count; specifier_index++)
   {
      specifier = [s objectAtIndex: specifier_index]; // quick access to current specifier

      // if this specifier has a name, try to localize it
      if ([specifier name] != NULL)
      {
         if ([[specifier name] isEqualToString: @"DEVICE_ID"])
            [specifier setName: [NSString stringWithFormat: @"%@: %d", LOCALIZE ([specifier name], [specifier name]), device_id]];
         else if ([[specifier name] isEqualToString: @"BUY_CODE"] && is_activated)
            [specifier setName: LOCALIZE (@"ALREADY_ACTIVATED", @"Already activated")];
         else
            [specifier setName: LOCALIZE ([specifier name], [specifier name])];
      }

      // is this specifier a dictionary ?
      if ([specifier titleDictionary])
      {
         // allocate and initialize a new dictionary that will contain the translated strings
         new_titles = [[NSMutableDictionary alloc] init];

         key_count = [[specifier titleDictionary] count];

         // for each entry in the original, non-translated dictionary...
         for (key_index = 0; key_index < key_count; key_index++)
         {
            key = [[specifier titleDictionary] objectAtIndex: key_index];
            [new_titles setObject: LOCALIZE ([[specifier titleDictionary] objectForKey: key], [[specifier titleDictionary] objectForKey: key]) forKey: key];
         }

         [specifier setTitleDictionary: [new_titles autorelease]];
      }
   }

   return (s);
}


- (id) navigationTitle
{
   // this function returns the localized page title for the preferences bundle

   // ask the localization API to hand us a replacement string for the bundle's title and return it
   return (LOCALIZE ([self title], [self title]));
}


- (id) getDaemonState: (PSSpecifier *) spec
{
   // this function returns YES if the Samba server is running, NO otherwise

   if (spec && Samba_IsDaemonRunning ())
      return ([NSNumber numberWithBool: YES]); // if Samba is running, return YES

   return ([NSNumber numberWithBool: NO]); // else return NO
}


- (void) setConfigValue: (id) value forSpecifier: (PSSpecifier *) spec
{
   UIAlertView *alert;

   // are we updating the activation code ?
   if (spec == [self specifierForID: @"activation_code"])
   {
      // were we already activated ?
      if (is_activated)
         return; // if so, just discard the change

      // else is it the correct code ?
      else if ((unsigned long) atol ([value UTF8String]) == activation_code)
      {
         is_activated = true;
         alert = [[UIAlertView alloc] initWithTitle: LOCALIZE (@"THANKYOU", @"Thank you") message: LOCALIZE (@"THANKYOU_TEXT", @"Thank you for registering this app. You may now enable the root share and access all files on your device.") delegate: self cancelButtonTitle: LOCALIZE (@"OK_BUTTON", @"OK") otherButtonTitles: nil];
         [alert show]; // and show it
      }

      // else it's NOT the correct code
      else
      {
         alert = [[UIAlertView alloc] initWithTitle: LOCALIZE (@"ACTIVATIONFAILURE", @"Activation failure") message: LOCALIZE (@"ACTIVATIONFAILURE_TEXT", @"The activation code you entered could not be validated.") delegate: self cancelButtonTitle: LOCALIZE (@"OK_BUTTON", @"OK") otherButtonTitles: nil];
         [alert show]; // and show it

         return; // most important, don't set the activation code at all
      }
   }

   // else are we toggling the root share on ?
   else if ((spec == [self specifierForID: @"share_fsroot_isenabled"]) && ((CFBooleanRef) value == kCFBooleanTrue))
   {
      // are we activated ?
      if (is_activated)
      {
         // prepare an alert box when trying to enable root filesystem sharing
         alert = [[UIAlertView alloc] initWithTitle: LOCALIZE (@"WARNING", @"Warning") message: LOCALIZE (@"WARNING_TEXT", @"You are now exposing the whole filesystem over the network.\n\nThis is a dangerous feature.\n\nYour device will be more exposed to potential intruders, and incorrect changes to system files may also render your device unusable.") delegate: self cancelButtonTitle: LOCALIZE (@"YES_BUTTON", @"Yes, I know") otherButtonTitles: nil];
         [alert show]; // and show it
      }

      // else we are NOT activated
      else
      {
         // prepare an alert box when trying to enable root filesystem sharing
         alert = [[UIAlertView alloc] initWithTitle: LOCALIZE (@"INFORMATION", @"Information") message: LOCALIZE (@"REGONLY_TEXT", @"To unlock full filesystem access, please register this application using an activation code.\n\nTo get your activation code, use the Activate button at the bottom of this settings page.") delegate: self cancelButtonTitle: LOCALIZE (@"OK_BUTTON", @"OK") otherButtonTitles: nil];
         [alert show]; // and show it

         return; // most important, don't toggle the share at all
      }
   }

   // else are we toggling a hidden share on ? if so, display a warning
   else if ((spec == [self specifierForID: @"share_storage_ishidden"]) && ((CFBooleanRef) value == kCFBooleanTrue))
   {
      alert = [[UIAlertView alloc] initWithTitle: LOCALIZE (@"INFORMATION", @"Information") message: [NSString stringWithFormat: @"%@\n\n\\\\%s\\%s$", LOCALIZE (@"HIDDENSHARE_TEXT", @"This share will no longer appear in Windows Explorer. As a hidden share, it can still be reached or added as a network drive at the following address:"), STRING_FOR_ID (@"netbios_name"), STRING_FOR_ID (@"share_storage_name")] delegate: self cancelButtonTitle: LOCALIZE (@"OK_BUTTON", @"OK") otherButtonTitles: nil];
      [alert show]; // and show it
   }
   else if ((spec == [self specifierForID: @"share_camera_ishidden"]) && ((CFBooleanRef) value == kCFBooleanTrue))
   {
      alert = [[UIAlertView alloc] initWithTitle: LOCALIZE (@"INFORMATION", @"Information") message: [NSString stringWithFormat: @"%@\n\n\\\\%s\\%s$", LOCALIZE (@"HIDDENSHARE_TEXT", @"This share will no longer appear in Windows Explorer. As a hidden share, it can still be reached or added as a network drive at the following address:"), STRING_FOR_ID (@"netbios_name"), STRING_FOR_ID (@"share_camera_name")] delegate: self cancelButtonTitle: LOCALIZE (@"OK_BUTTON", @"OK") otherButtonTitles: nil];
      [alert show]; // and show it
   }
   else if ((spec == [self specifierForID: @"share_fsroot_ishidden"]) && ((CFBooleanRef) value == kCFBooleanTrue))
   {
      alert = [[UIAlertView alloc] initWithTitle: LOCALIZE (@"INFORMATION", @"Information") message: [NSString stringWithFormat: @"%@\n\n\\\\%s\\%s$", LOCALIZE (@"HIDDENSHARE_TEXT", @"This share will no longer appear in Windows Explorer. As a hidden share, it can still be reached or added as a network drive at the following address:"), STRING_FOR_ID (@"netbios_name"), STRING_FOR_ID (@"share_fsroot_name")] delegate: self cancelButtonTitle: LOCALIZE (@"OK_BUTTON", @"OK") otherButtonTitles: nil];
      [alert show]; // and show it
   }

   // update the settings bundle specifiers and save them
   [self setPreferenceValue: value specifier: spec];
   [[NSUserDefaults standardUserDefaults] synchronize];

   // update Samba config and restart daemon if needed
   Samba_Setup (&samba, BOOL_FOR_ID (@"enable_sharing"), STRING_FOR_ID (@"workgroup"), STRING_FOR_ID (@"netbios_name"), STRING_FOR_ID (@"description"), STRING_FOR_ID (@"password"));
   Samba_SetupShare (&samba.storage, BOOL_FOR_ID (@"share_storage_isenabled"), STRING_FOR_ID (@"share_storage_name"), "/var/samba/storage", STRING_FOR_ID (@"share_storage_description"), BOOL_FOR_ID (@"share_storage_isreadonly"), BOOL_FOR_ID (@"share_storage_ishidden"));
   Samba_SetupShare (&samba.camera, BOOL_FOR_ID (@"share_camera_isenabled"), STRING_FOR_ID (@"share_camera_name"), "/var/mobile/Media/DCIM/100APPLE", STRING_FOR_ID (@"share_camera_description"), BOOL_FOR_ID (@"share_camera_isreadonly"), BOOL_FOR_ID (@"share_camera_ishidden"));
   Samba_SetupShare (&samba.fsroot, is_activated && BOOL_FOR_ID (@"share_fsroot_isenabled"), STRING_FOR_ID (@"share_fsroot_name"), "/", STRING_FOR_ID (@"share_fsroot_description"), BOOL_FOR_ID (@"share_fsroot_isreadonly"), BOOL_FOR_ID (@"share_fsroot_ishidden"));
   Samba_WriteConfig ();

   return; // finished
}


- (void) pushButton: (PSSpecifier *) spec
{
   // are we hitting the buy button ? if so, fire up Safari
   if ((spec == [self specifierForID: @"activation_button"]) && !is_activated)
      [[UIApplication sharedApplication] openURL: [NSURL URLWithString: [NSString stringWithFormat: @"http://www.pmbaty.com/iosbuildenv/pcnetworking/index.php?id=%d", device_id]]];

   return; // finished
}


@end


static void ComputeDeviceIDAndActivationCode (const char *unique_id)
{
   // undisclosed.
}


static bool Samba_IsDaemonRunning (void)
{
   // this function returns TRUE if Samba is running, FALSE otherwise

   FILE *fp;

   // try to open the daemon's pid file
   if ((fp = fopen (SMBD_PIDFILE, "rb")) != NULL)
   {
      fclose (fp); // if we can, then it's running
      return (true); // so close it and return TRUE
   }

   return (false); // if pid file not found, Samba isn't running, return FALSE
}


static void Samba_Setup (sambasettings_t *settings, bool is_running, const char *workgroup, const char *netbios_name, const char *description, const char *password)
{
   // handy wrapper to populate a sambasettings_t structure

   settings->is_running = is_running;
   strncpy (settings->workgroup, workgroup, sizeof (settings->workgroup));
   settings->workgroup[sizeof (settings->workgroup) - 1] = 0;
   strncpy (settings->netbios_name, netbios_name, sizeof (settings->netbios_name));
   settings->netbios_name[sizeof (settings->netbios_name) - 1] = 0;
   strncpy (settings->description, description, sizeof (settings->description));
   settings->description[sizeof (settings->description) - 1] = 0;
   strncpy (settings->password, password, sizeof (settings->password));
   settings->password[sizeof (settings->password) - 1] = 0;

   return; // finished
}


static void Samba_SetupShare (sambashare_t *share, bool is_enabled, const char *name, const char *path, const char *description, bool is_readonly, bool is_hidden)
{
   // handy wrapper to populate a sambashare_t structure

   share->is_enabled = is_enabled;
   strncpy (share->name, name, sizeof (share->name));
   share->name[sizeof (share->name) - 1] = 0;
   strncpy (share->path, path, sizeof (share->path));
   share->path[sizeof (share->path) - 1] = 0;
   strncpy (share->description, description, sizeof (share->description));
   share->description[sizeof (share->description) - 1] = 0;
   share->is_readonly = is_readonly;
   share->is_hidden = is_hidden;

   return; // finished
}


static void Samba_WriteConfig (void)
{
   // this function updates Samba's configuration and restarts the daemon if needed

   FILE *fp;
   int length;
   int index;

   // set the machine name uppercase
   length = strlen (samba.netbios_name);
   for (index = 0; index < length; index++)
      samba.netbios_name[index] = toupper (samba.netbios_name[index]);
   samba.netbios_name[index] = 0;

   // set the workgroup name uppercase
   length = strlen (samba.workgroup);
   for (index = 0; index < length; index++)
      samba.workgroup[index] = toupper (samba.workgroup[index]);
   samba.workgroup[index] = 0;

   // update the shares config file
   if ((fp = fopen (SHARES_CONFIGFILE, "wb")) != NULL)
   {
      fprintf (fp, "[global]\n");
      fprintf (fp, "workgroup     = %s\n", samba.workgroup);
      fprintf (fp, "netbios name  = %s\n", samba.netbios_name);
      fprintf (fp, "server string = \"%s\"\n", samba.description);
      if (samba.storage.is_enabled)
      {
         fprintf (fp, "\n");
         fprintf (fp, "[%s%s]\n", samba.storage.name, (samba.storage.is_hidden ? "$" : ""));
         fprintf (fp, "comment = %s\n", samba.storage.description);
         fprintf (fp, "path = %s\n", samba.storage.path);
         fprintf (fp, "fstype = Apple\n");
         fprintf (fp, "browsable = yes\n");
         fprintf (fp, "writable = %s\n", (samba.storage.is_readonly ? "no" : "yes"));
         fprintf (fp, "guest only = yes\n");
         fprintf (fp, "guest ok = yes\n");
      }
      if (samba.camera.is_enabled)
      {
         fprintf (fp, "\n");
         fprintf (fp, "[%s%s]\n", samba.camera.name, (samba.camera.is_hidden ? "$" : ""));
         fprintf (fp, "comment = %s\n", samba.camera.description);
         fprintf (fp, "path = %s\n", samba.camera.path);
         fprintf (fp, "fstype = Apple\n");
         fprintf (fp, "browsable = yes\n");
         fprintf (fp, "writable = %s\n", (samba.camera.is_readonly ? "no" : "yes"));
         fprintf (fp, "guest only = yes\n");
         fprintf (fp, "guest ok = yes\n");
      }
      if (is_activated && samba.fsroot.is_enabled)
      {
         fprintf (fp, "\n");
         fprintf (fp, "[%s%s]\n", samba.fsroot.name, (samba.fsroot.is_hidden ? "$" : ""));
         fprintf (fp, "comment = %s\n", samba.fsroot.description);
         fprintf (fp, "path = %s\n", samba.fsroot.path);
         fprintf (fp, "fstype = Apple\n");
         fprintf (fp, "browsable = yes\n");
         fprintf (fp, "writable = %s\n", (samba.fsroot.is_readonly ? "no" : "yes"));
         fprintf (fp, "guest only = yes\n");
         fprintf (fp, "guest ok = yes\n");
      }
      fclose (fp);
   }

   // update the passwords file
   if ((fp = fopen (PASSWORDS_CONFIGFILE, "wb")) != NULL)
   {
      fprintf (fp, "%s\n", samba.password);
      fprintf (fp, "%s\n", samba.password);
      fclose (fp);
   }

   // stop the daemon, and eventually restart it if needed
   notify_post (NOTIFICATIONNAME_STOP);
   if (samba.is_running)
      notify_post (NOTIFICATIONNAME_START);

   return; // finished, Samba config has been updated
}
